module vga (clk,r,g,b,hs,vs,vgaclk,       // vga interface
             addr,datain,dataout,wr,ready,      // cpu interface
             sram_addr,sram_ce_n,sram_dq,sram_lb_n,sram_oe_n,sram_ub_n,sram_we_n);// sram interface); 
				 // 1440x900 160x50 char with border text mode vga driver for DE2-115

// ver 0.14 alpha --- 2014.08.22
// (c) Piotr Kardasz pik33@we.pb.edu.pl
// This is free software, license: LGPL v.2.0 or newer

output reg [7:0] r,g,b;
output reg hs,vs;
output reg vgaclk; 
input wire clk;  //input clock=2x vgaclk
output wire ready;

reg e,f,mclk;
reg [7:0] char2,char,attr,attr2,ramdata,ramdata2,brr,brb,brg,cposx,cposy;
reg [19:0] caddr,caddr2,cpos;
reg [9:0] lcnt;
reg [0:2] lstate,pstate; // 00 vis, 01 fp, 10 sync, 11 bp 100 border
reg [10:0] bb,c,d,ramaddr,pcnt;
reg [31:0] fcnt; //frame counter
reg [15:0] start;

input wire [19:0] addr;
input wire [15:0] datain;
output reg [15:0] dataout;
input wire wr;
output reg [19:0] sram_addr;
inout wire [15:0] sram_dq;
output reg sram_ce_n;
output reg sram_lb_n;
output reg sram_oe_n;
output reg sram_ub_n;
output reg sram_we_n;
reg [15:0] sramdqout,datain2;
reg [19:0] addr2;
reg fl;
reg oldwr;


(* ramstyle = "M9K" *) reg [7:0] fontdef[2048];  // character generator

initial
  begin
  $readmemh("st4font.hex",fontdef);
  start=20'b0000_0000_0000_0000_0000;
  fl<=0;
  end
  
assign sram_dq = (sram_we_n==0) ? sramdqout : 16'bz;
assign ready=~fl;

always
  begin
  sram_lb_n=0;
  sram_ub_n=0;
  sram_ce_n=0;
  sram_oe_n=0;
  end

  
always@(posedge clk)

  begin
  f<=~f; // phase
  if (f==0)
    begin


//------------- sram ---------------------------------
    case ((pcnt+1)&7)

	 0,1,2: 
	    begin
	    attr<=attr2;
		 if (fl==1)
	      begin  
	      sramdqout<=datain2;
	      sram_addr<=addr2;
	      if (((pcnt+1)&7)>0) sram_we_n<=0; else sram_we_n<=1;
	      end
		 else
		   begin
	      sram_addr<=addr;
	      sram_we_n<=1;
	      end
		 end	
	 3: begin	
	    dataout<=sram_dq;
    	 sram_we_n<=1;
		 end
	 4,5,6: begin
	    fl<=0;	 
       sram_addr<=caddr;
       end
	 7: begin
	     {char,attr2}<=sram_dq;
        if (wr==1) 
	       begin 
          addr2<=addr; 
	       datain2<=datain; 
	       fl<=1;
          end
		  end	 
		endcase  
	      		 
// ------------------- end sram -----------------------------  

  	   pcnt<=pcnt+1;
       // line and pixel counters & state determination
	   if (pcnt==231) pstate<=4; // left border
	   if (pcnt==311) pstate<=0; // display
	   if (pcnt==1591) pstate<=4; // right border
	   if (pcnt==1671) pstate<=1; // front porch
	   if (pcnt==1751) begin pstate<=2; hs<=0; end// hsync
	   if (pcnt==1903)           // back porch 
	     begin 
		  pstate<=3; 
        hs<=1;
		  pcnt<=0;
		  lcnt<=lcnt+1; 
		  if (lcnt==27) lstate<=4;   // top border
		  if (lcnt==77) lstate<=0;   // display
		  if (lcnt==877) lstate<=4;  // bottom border
		  if (lcnt==927) lstate<=1;  // front porch v
		  if (lcnt==928) begin lstate<=2; vs<=1; end // vsync 
		  if (lcnt==931) begin lstate<=3; vs<=0; lcnt<=0; fcnt<=fcnt+1; end // back porch v
		  end
 
      caddr<=((lcnt-78)>>4)*160+((pcnt-303)>>3); // compute char addr in fb
      caddr2<=((lcnt-78)>>4)*160+((pcnt-311)>>3);
      if (lstate==2) //vblank
	     begin //stub!! 
	     caddr<=20'hFFF60+((pcnt-311)>>3);
	     if (caddr==20'hFFFF0) brr<=attr2; // get border colors
	     if (caddr==20'hFFFF1) brg<=attr2;
        if (caddr==20'hFFFF2) brb<=attr2;
		  if (caddr==20'hFFFF3) begin cposx<=char; cposy<=attr2; end
		   cpos<=160*cposy+cposx;
        end
    
      else if (((lstate==4)&&((pstate&3)==0)) || ((pstate==4)&&(lstate==0))) // borders
        begin 
        r<=brr; g<=brg; b<=brb;	
        bb<=(lcnt-78) & 15;
	     c<=(pcnt+1) & 7;
		  ramaddr<=((char<<4)|bb); 
        end		 
	 
      else if (lstate==0 && pstate==0) // active display
        begin
	     bb<=(lcnt-78) & 15;
	     c<=(pcnt+1) & 7;
		  ramaddr<=((char<<4)|bb); 
	     if ((cpos==caddr2)&& (bb>12) && (fcnt[4]==1))
		    begin 
		    e<=~ramdata[c]; 
          if (e==1) begin r<=192; g<=192; b<=192; end else begin r<=0; b<=0; g<=0; end
		    end
        else 
          begin
          e<=ramdata[c];
          if (e==1) r<=192*attr[2]+63*attr[3]; else r<=192*attr[6]+63*attr[7];
          if (e==1) g<=192*attr[1]+63*attr[3]; else g<=192*attr[5]+63*attr[7];
          if (e==1) b<=192*attr[0]+63*attr[3]; else b<=192*attr[4]+63*attr[7];
		    end
		  end
	   else 
        begin r<=0; g<=0; b<=0; end		
      end 
    vgaclk<=f; // vga dac clock output 

  end
 
always@ (negedge clk) 
  begin
  ramdata=fontdef[ramaddr];
  end 

endmodule 